
/* tiny cross-platform quick and dirty snippet demo for parsing command line arguments
 * app key1:value1 key2:value2 ...
 *
 * For compilation:
 *  - on UNIX-like (GNU/Linux, BSD, OSX, etc) : cc parser.c -o parser
 *  - on Windows : cl parser.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BUFFER_SIZE 4096

struct vars {
	char name[BUFFER_SIZE];
	char host[BUFFER_SIZE];
	char port[BUFFER_SIZE];
	int verbose;
};

int main(int argc, const char *argv[])
{
	const char *options[] = { "host", "port", NULL };
	
	int opt_port = 0;
	int opt_host = 0;
	
	struct vars values;
	
	unsigned int delim = ':';
	char *zbuffer;
	char *value;
	const char *str;
	const char *ch;
	
	int c = 0;
	int i = 0;
	zbuffer = (char *)malloc(BUFFER_SIZE);
	value = (char *)malloc(BUFFER_SIZE);
	
	memset(zbuffer, '\0', BUFFER_SIZE);
	memset(values.name, '\0', BUFFER_SIZE);
	memset(values.host, '\0', BUFFER_SIZE);
	memset(values.port, '\0', BUFFER_SIZE);
	values.verbose = 0;
	
	strcpy(values.name, argv[0]);
	
	while(*++argv) {
		if (strcmp(*argv, "verbose") == 0) {
			values.verbose = 1;
			c++;
		}
		for (i = 0; options[i]; i++) {
			str = strstr(*argv, options[i]);
			if (!str) {
				continue;
			}
			
			ch = strchr(*argv, delim);
			if (!ch) {
				continue;
			}
			
			memset(value, '\0', BUFFER_SIZE);
			strcpy(value, ++ch);
			
			printf("key:   %s\n", options[i]);
			printf("value: %s\n", value);
			
			switch (i) {
			case 0:
				printf("host\n");
				strcpy(values.host, value);
				c++;
				break;
			case 1:
				printf("port\n");
				if ((memcmp(values.port, zbuffer, BUFFER_SIZE) == 0) && (strlen(value) > 0)) {
					strcpy(values.port, value);
					opt_port = 1;
					c++;
				} else {
					printf("parsing error: see usage\n");
					return 0;
				}
				break;
			default:
				break;
			}
		}
	}
	
	printf("--------------\n");
	printf("%d(%d) : %d\n", argc, argc-1, c);
	if (c != argc - 1) {
		printf("parsing error: arguments count: see usage\n");
	}
	printf("parsing result: \n");
	printf("name = %s\n", values.name);
	printf("host = %s\n", values.host);
	printf("port = %s\n", values.port);
	printf("verbose = %d\n", values.verbose);
	
	if (!opt_port) {
		printf("No value for 'port' key\n");
	}
	
	return 0;
}



