
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>

#define NAME "netcmd"
#define INVOKE "netcmd > "

#define MOTD printf("\tWelcome!\n"); \
		system("uname -s"); \
		system("uname -n"); \
		system("uname -r"); \
		system("uname -v"); \
		system("uname -m"); \
		system("uname -o"); \
		printf("\n"); \
		system("w"); \
		printf("\n"); \
		system("id"); \
		printf("\n");

/* number of pending connections in the queue */
#define BACKLOG 10
/* max number of bytes for transfer at once */
#define MAXDATASIZE 32*1024

/* usage info */
int usage(void)
{
	printf("%s [port]\n", NAME);
	return 1;
}

/* signal handler */
void sigchld_handler(int s)
{
	while(waitpid(-1, NULL, WNOHANG) > 0);
}

/* get sockaddr : IPv4 or IPv6 */
void *get_in_addr(struct sockaddr *sa)
{
	if (sa->sa_family == AF_INET) {
		return &(((struct sockaddr_in*)sa)->sin_addr);
	}
	
	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

int shell_server(const char *port)
{
	/* sockfd for listen(), new_fd for accept() */
	int sockfd, new_fd;
	
	/* client's address information */
	struct sockaddr_storage their_addr;
	
	struct sigaction sa;
	struct addrinfo hints, *servinfo, *p;
	socklen_t sin_size;
	char s[INET6_ADDRSTRLEN];
	
	char cmd[MAXDATASIZE];
	int numbytes;
	int rv;
	int yes = 1;
	
	/* init routine */
	memset(&hints, 0, sizeof hints);
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_family = AF_UNSPEC;
	hints.ai_flags = AI_PASSIVE;
	
	if ((rv = getaddrinfo(NULL, port, &hints, &servinfo)) != 0) {
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
		return 1;
	}
	
	/* loop through all the results and bind to the first we can */
	for (p = servinfo; p != NULL; p = p->ai_next) {
		if ((sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
			perror("server: socket");
			continue;
		}
		
		if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
			perror("setsockopt");
			exit(1);
		}
		
		if (bind(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
			close(sockfd);
			perror("server: bind");
			continue;
		}
		
		break;
	}
	
	if (p == NULL)  {
		fprintf(stderr, "server: failed to bind\n");
		return 2;
	}
	
	/* free structure since it's not using anymore */
	freeaddrinfo(servinfo);
	
	if (listen(sockfd, BACKLOG) == -1) {
		perror("listen");
		exit(1);
	}
	
	/* reap all dead processes */
	sa.sa_handler = sigchld_handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa, NULL) == -1) {
		perror("sigaction");
		exit(1);
	}
	
	/* duplicate desciptors for backup */
	int std_out = dup(STDOUT_FILENO);
	int std_err = dup(STDERR_FILENO);
	
	
	printf("server: waiting for connections...\n");
	
	/* int address = 0x7F000001; */
	
	/* main accept() loop */
	while (1) {
		sin_size = sizeof their_addr;
		new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);
		if (new_fd == -1) {
			perror("accept");
			continue;
		}
		
		inet_ntop(their_addr.ss_family, get_in_addr((struct sockaddr *)&their_addr), s, sizeof s);
		
		printf("server: got connection from %s\n", s);
		
		/* source address checking */
		/*
		if (address != ntohl(((struct sockaddr_in *)&their_addr)->sin_addr.s_addr)) {
			printf("I won't talk with you\n");
			exit(3);
		}
		*/
		
		/* init the child process for processing connect() from client */
		if (!fork()) {
			/* close parents' socket for accept() */
			close(sockfd);
			
			/* redirect standard output/error to the file using dup2 */
			if (dup2(new_fd, STDOUT_FILENO) < 0) {
				printf("dup2: out\n");
				return 1;
			}
			if (dup2(new_fd, STDERR_FILENO) < 0) {
				printf("dup2: err\n");
				return 1;
			}
			/*
			printf("Welcome to:\n");
			system("uname -a");
			*/
			MOTD
			send(new_fd, INVOKE, strlen(INVOKE), 0);
			while (((numbytes = recv(new_fd, cmd, MAXDATASIZE-1, 0)) != -1) && numbytes != 0) {
				cmd[numbytes] = '\0';
				system(cmd);
				send(new_fd, INVOKE, strlen(INVOKE), 0);
			}
			
			close(new_fd);
			
			/* flush descriptors so write data before restoring output */
			fflush(stdout);
			fflush(stderr);
			
			/* restore output */
			dup2(std_out, STDOUT_FILENO);
			dup2(std_err, STDERR_FILENO);
			
			printf("Connection closed.\n");
			
			exit(0);
		}
		
		/* socket has been processed in the child process */
		close(new_fd);
	}
	
	return 0;
}

int main(int argc, const char *argv[])
{
	if (argc != 2) {
		return usage();
	}
	
	if ((atoi(argv[1]) < 1024) && (getuid())) {
		printf("You need to be root\n");
		return -1;
	}
	
	shell_server(argv[1]);
	
	return 0;
}

