
/* includes */

#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <iostream.h>

/* data types */

#ifdef _MSC_VER
// The following two structures need to be packed tightly, but unlike
// Borland C++, Microsoft C++ does not do this by default.
#pragma pack(1)
#endif

/* exist
struct timeval {
	long tv_sec;
	long tv_usec;
};
*/

/* exist
struct in_addr {
    unsigned long s_addr;
};
*/

/* ip packet header */
struct ip {
	unsigned int ip_hl:4;       /* header length */
	unsigned int ip_v:4;       /* version */
	unsigned char ip_tos;       /* type of service */
	unsigned short int ip_len;  /* total length */
	unsigned short int ip_id;   /* identification */
	unsigned short int ip_off;  /* fragment offset field / flags */
/*
#define	IP_RF 0x8000                // reserved fragment flag
#define	IP_DF 0x4000                // dont fragment flag
#define	IP_MF 0x2000                // more fragments flag
#define	IP_OFFMASK 0x1fff           // mask for fragmenting bits
*/
	unsigned char ip_ttl;       /* time to live */
	unsigned char ip_p;         /* protocol */
	unsigned short int ip_sum;  /* checksum */
	struct in_addr ip_src;      /* source address */
	struct in_addr ip_dst;      /* dest address */

};

/* icmp packet header - pretty not standart */
struct icmp {
	unsigned char icmp_type;          /* type of message */
	unsigned char icmp_code;          /* type sub code */
	unsigned short int icmp_cksum;    /* checksum */
	unsigned short int icmp_id;       /* identification */
	unsigned short int icmp_seq;      /* sequence number */
	unsigned long int icmp_tmstmp;    /* not part of ICMP - used for time */
};

#ifdef _MSC_VER
#pragma pack()
#endif

/* defines */

#define WIN32_LEAN_AND_MEAN

/* icmp packet types */
#define ICMP_ECHOREPLY         0  /* Echo Reply */
#define ICMP_DEST_UNREACH      3  /* Destination Unreachable */
#define ICMP_SOURCE_QUENCH     4  /* Source Quench */
#define ICMP_REDIRECT          5  /* Redirect (change route) */
#define ICMP_ECHO              8  /* Echo Request */
#define ICMP_TIME_EXCEEDED    11  /* Time Exceeded */
#define ICMP_PARAMETERPROB    12  /* Parameter Problem */
#define ICMP_TIMESTAMP        13  /* Timestamp Request */
#define ICMP_TIMESTAMPREPLY   14  /* Timestamp Reply */
#define ICMP_INFO_REQUEST     15  /* Information Request */
#define ICMP_INFO_REPLY       16  /* Information Reply */
#define ICMP_ADDRESS          17  /* Address Mask Request */
#define ICMP_ADDRESSREPLY     18  /* Address Mask Reply */

/* minimum ICMP packet size */
#define ICMP_MINLEN            8

/* additional values */
#define DEFAULT_PACKET_SIZE   32
#define DEFAULT_TTL           30
#define MAX_PING_DATA_SIZE  1024
#define MAX_PING_PACKET_SIZE  (MAX_PING_DATA_SIZE + sizeof(ip))

#define BUFSIZE 3000

/* global variables */
SOCKET sd;                      /* socket descriptor */
unsigned short int pid;      /* PID */
struct sockaddr_in servaddr; /* sockaddr() for sending   */
struct sockaddr_in from;     /* sockaddr() for receiving */

double tmin = 999999999.0;   /* minimal round-trip time */
double tmax = 0;             /* maximum round-trip time */
double tsum = 0;             /* summary time for calculating average time */
int nsent = 0;               /* number of sended packages */
int nreceived = 0;           /* number of received packages */

/* additional macros */
#ifndef min
#define min(x, y) ((x<y)? (x): (y))
#endif

#ifndef max
#define max(x, y) ((x>y)? (x): (y))
#endif

/* functions */

unsigned short in_cksum(unsigned short *addr, int len)
{
	unsigned short result;
	unsigned int sum = 0;
	
	/* sum for all dwords */
	while (len > 1) {
		sum += *addr++;
		len -= 2;
	}
	
	/* add byte to sum */
	if (len == 1) {
		sum += *(unsigned char*) addr;
	}
	
	sum = (sum >> 16) + (sum & 0xFFFF); /* carrying */
	sum += (sum >> 16);                 /* again */
	result = ~sum;                      /* and invert */
	return result;
}

void pinger(void)
{
	int icmplen;
	struct icmp *icmp;
	char sendbuf[BUFSIZE];
	
	icmp = (struct icmp *) sendbuf;
	
	/* prepare ICMP fields for package */
	icmp->icmp_type = ICMP_ECHO;
	icmp->icmp_code = 0;
	icmp->icmp_id = pid;
	icmp->icmp_seq = nsent++;
	icmp->icmp_tmstmp = GetTickCount();
	
	/* length of ICMP header: 8(control) + 56(data) */
	icmplen = 8 + 56;
	/* check sum of ICMP header and data */
	icmp->icmp_cksum = 0;
	icmp->icmp_cksum = in_cksum((unsigned short *) icmp, icmplen);
	
	if (sendto(sd, sendbuf, icmplen, 0, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0) {
		printf("error: sendto() failed\n");
		return;
	}
	
	return;
}

void output(char *ptr, int len, unsigned long tvrecv)
{
	int iplen;
	int icmplen;
	struct ip *ip;
	struct icmp *icmp;
	//struct timeval *tvsend;
	unsigned long tvsend;
	double rtt;
	
	ip = (struct ip *) ptr; /* beginning of IP header */
	iplen = ip->ip_hl << 2; /* length of IP header */
	
	icmp = (struct icmp *) (ptr + iplen); /* beginning of ICMP header */
	if ( (icmplen = len - iplen) < 8) {   /* length of ICMP header */
		fprintf(stderr, "icmplen (%d) < 8", icmplen);
	}
	
	if (icmp->icmp_type == ICMP_ECHOREPLY) {
		if (icmp->icmp_id != pid) {
			/* this is ECHO REPLY for another process */
			return;
		}
		
		tvsend = icmp->icmp_tmstmp;
		//tvsend = (struct timeval *) icmp->icmp_data;
		//tv_sub(tvrecv, tvsend);
		
		/* round-trip time */
		rtt = tvrecv - tvsend; //tvrecv->tv_sec * 1000.0 + tvrecv->tv_usec / 1000.0;
		
		nreceived++;
		
		tsum += rtt;
		if (rtt < tmin) {
			tmin = rtt;
		}
		if (rtt > tmax) {
			tmax = rtt;
		}
		
		printf("%d bytes from %s: icmp_seq=%u, ttl=%d, time=%.3f ms\n", icmplen, inet_ntoa(from.sin_addr), icmp->icmp_seq, ip->ip_ttl, rtt);
	}
	
	return
}

int main(int argc, char *argv[])
{
	int size;
	socklen_t fromlen;
	int n;
	struct timeval tval;
	unsigned long tmstmp;
	char recvbuf[BUFSIZE];
	struct hostent *hp;
	const int on = 1;
	
	if (argc != 2) {
		fprintf(stderr, "Usage: %s <hostname>\n", argv[0]);
		exit(-1);
	}
	
	WSAData wsaData;
	if (WSAStartup(MAKEWORD(2, 1), &wsaData) != 0) {
		printf("error: failed to load WINSOCK library\n");
		return 1;
	}
	
	pid = (unsigned short int) GetCurrentProcessId();
	
	if ((hp = gethostbyname(argv[1])) == 0) {
		printf("error: gethostbyname() failed\n");
		return 1;
	}
	
	if ((sd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) == INVALID_SOCKET) {
		printf("error: socket() failed\n");
		return 1;
	}
	
	if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, (char *) &on, sizeof(on)) == SOCKET_ERROR) {
		printf("error: setsockopt() for broadcast failed");
		return 1;
	}
	
	size = 60 *1024;
	if (setsockopt(sd, SOL_SOCKET, SO_RCVBUF, (char *) &size, sizeof(size)) == SOCKET_ERROR) {
		printf("error: setsockopt() for rcvbuf failed");
		return 1;
	}
	
	//bzero(&servaddr, sizeof(servaddr));
	memset(&servaddr, '\0', sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr= *((struct in_addr *) hp->h_addr);
	
	fromlen = sizeof(from);
	
	/* */
	
	pinger();
	
	n = recvfrom(sd, recvbuf, sizeof(recvbuf), 0, (struct sockaddr *)&from, &fromlen);
	
	if (n < 0) {
		printf("error: recvfrom() failed\n");
		return 1;
	}
	
	/* get current system time */
	tmstmp = GetTickCount();
	
	/* callback function for parcing output */
	output(recvbuf, n, tmstmp);
	
	/* */
	
	WSACleanup();
	
	return 0;
}
