#!/bin/bash

export LOGF="/home/pi/fishing.log"

usbid="1-1.2.2"
iface_wifi=wlan1
iface_inet=ppp0
iface_nmid=Megafon
timeout=60
hostapath="/home/pi/tools/hostapd-1.0-karma/hostapd"
apconf="/home/pi/tools/hostapd-1.0-karma/ap.conf"
logpath="/home/pi/tools/hostapd-1.0-karma/logs"
dhcplog="/var/log/syslog"
gsmline="GSM modem (1-port) converter now attached"
stmp="`date +%Y.%m.%d_%H.%M.%S`"

LOG()
{
	echo `date "+%Y.%m.%d %H:%M:%S"` msg:${@} | tee -a "${LOGF}"
}

modules()
{
	LOG " starting ppp modules ..."
	
	modprobe ppp_deflate
	modprobe bsd_comp
	modprobe ppp_async
	modprobe ppp_generic
}

dhcp()
{
	service isc-dhcp-server ${@}
}

modem()
{
	LOG " waiting internet connection ..."
	
	rtrn=1
	
	while [ "${rtrn}" -ne 0 ]; do
		LOG " resetting modem ..."
		echo 0 > /sys/bus/usb/devices/${usbid}/bConfigurationValue
		sleep 5
		echo 1 > /sys/bus/usb/devices/${usbid}/bConfigurationValue
		sleep ${timeout}
		LOG " trying to connect ..."
		nmcli con up id ${iface_nmid}
		rtrn="${?}"
	done;
}

iflocal()
{
	LOG " starting network interface ..."
	
	#ifconfig ${iface_wifi} down
	ifconfig ${iface_wifi} up 10.0.0.1 netmask 255.255.255.0
}

nat()
{
	LOG " starting internet sharing ..."
	
	iptables --flush
	iptables --table nat --flush
	iptables --delete-chain
	iptables --table nat --delete-chain
	iptables --table nat --append POSTROUTING --out-interface ${iface_inet} -j MASQUERADE
	iptables --append FORWARD --in-interface ${iface_wifi} -j ACCEPT
	
	#Uncomment the line below if facing problems while sharing PPPoE
	#iptables -I FORWARD -p tcp --tcp-flags SYN,RST SYN -j TCPMSS --clamp-mss-to-pmtu
	
	sysctl -w net.ipv4.ip_forward=1
}

hostap()
{
	LOG " starting wireless access point ..."
	
	${hostapath}/hostapd -B -t -dd -f ${logpath}/host_${stmp} ${apconf}
	sleep 1
	LOG " starting karma ..."
	
	${hostapath}/hostapd_cli -i ${iface_wifi} karma_enable
}

dump()
{
	LOG " starting tcpdump ..."
	
	tcpdump -i ${iface_wifi} -w ${logpath}/dump_${stmp} -nn -XX -vvv -s0 -S -e &
}

connection()
{
	rtrn=0
	while [ "${rtrn}" -eq 0 ]; do
		sleep 60
		ifconfig ${iface_inet} > /dev/null 2>&1
		rtrn="${?}"
	done;
	
	# LOG " no internet connection: REBOOTING ..."
	LOG " no internet connection: RETRYING ..."
	
	#sleep 1
	#sync
	#sleep 1
	#reboot
	
	modem
}

catching()
{
	tail -f "${dhcplog}" | while read dhcpline; do
		line="`echo "${dhcpline}" | grep DHCPACK`"
		if [ -n "${line}" ]; then
			fish `echo "${line}" | awk '{print $8, $10, $11}' | sed 's,[(,)],,g'`
		fi
		# detecting when clients leave
		release="`echo "${dhcpline}" | grep "IEEE 802.11: disassociated"`"
		if [ -n "${release}" ]; then
			echo " <<< " `date "+%Y.%m.%d %H:%M:%S"`  "----"  `echo "${release}" | awk '{print $8}'`  "----" | tee -a "${logpath}"/fish_${stmp}
		fi
	done;
}

fish()
{
	# main callback for clients
	echo " >>> " `date "+%Y.%m.%d %H:%M:%S"`  ${1}  ${2}  ${3} | tee -a "${logpath}"/fish_${stmp}
}

LOG "==========================="
LOG " starting log ..."
LOG "==========================="

LOG " starting fishing ..."

LOG " stopping dhcp server ..."

dhcp stop

modules
sleep 5
# modem

iflocal
sleep 5
nat
sleep 5
dhcp start
sleep 5
hostap

LOG " restarting local wireless connection ..."

iflocal
sleep 5
dhcp start
sleep 1

dump &

LOG " happy fishing ..."

modem

connection &

echo "    TIME        IP        MAC        HOSTID    " | tee -a "${logpath}"/fish_${stmp}

catching

wait
